from inspect import isclass
from typing import Any, Dict, ItemsView, List, Optional, Type, Union, cast

from pydantic import validate_arguments

from starlite.controller import Controller
from starlite.enums import HttpMethod
from starlite.exceptions import ImproperlyConfiguredException
from starlite.handlers import (
    ASGIRouteHandler,
    BaseRouteHandler,
    HTTPRouteHandler,
    WebsocketRouteHandler,
)
from starlite.provide import Provide
from starlite.response import Response
from starlite.routes import ASGIRoute, BaseRoute, HTTPRoute, WebSocketRoute
from starlite.types import (
    AfterRequestHandler,
    BeforeRequestHandler,
    ControllerRouterHandler,
    Guard,
    ResponseHeader,
)
from starlite.utils import find_index, join_paths, normalize_path, unique


class Router:
    __slots__ = (
        "after_request",
        "before_request",
        "dependencies",
        "guards",
        "owner",
        "path",
        "tags",
        "response_class",
        "response_headers",
        "routes",
    )

    @validate_arguments(config={"arbitrary_types_allowed": True})
    def __init__(
        self,
        *,
        path: str,
        route_handlers: List[ControllerRouterHandler],
        tags: Optional[List[str]] = None,
        dependencies: Optional[Dict[str, Provide]] = None,
        guards: Optional[List[Guard]] = None,
        response_class: Optional[Type[Response]] = None,
        response_headers: Optional[Dict[str, ResponseHeader]] = None,
        # connection-lifecycle hook handlers
        before_request: Optional[BeforeRequestHandler] = None,
        after_request: Optional[AfterRequestHandler] = None,
    ):
        self.owner: Optional["Router"] = None
        self.routes: List[BaseRoute] = []
        self.path = normalize_path(path)
        self.tags = tags
        self.response_class = response_class
        self.dependencies = dependencies
        self.response_headers = response_headers
        self.guards = guards
        self.before_request = before_request
        self.after_request = after_request
        for route_handler in route_handlers or []:
            self.register(value=route_handler)

    @property
    def route_handler_method_map(self) -> Dict[str, Union[WebsocketRouteHandler, Dict[HttpMethod, HTTPRouteHandler]]]:
        """
        Returns dictionary that maps paths (keys) to a list of route handler functions (values)
        """
        route_map: Dict[str, Union[WebsocketRouteHandler, Dict[HttpMethod, HTTPRouteHandler]]] = {}
        for route in self.routes:
            if isinstance(route, HTTPRoute):
                if not isinstance(route_map.get(route.path), dict):
                    route_map[route.path] = {}
                for route_handler in route.route_handlers:
                    for method in route_handler.http_methods:
                        route_map[route.path][method] = route_handler  # type: ignore
            else:
                route_map[route.path] = cast(WebSocketRoute, route).route_handler
        return route_map

    @staticmethod
    def map_route_handlers(
        value: Union[Controller, BaseRouteHandler, "Router"],
    ) -> ItemsView[str, Union[WebsocketRouteHandler, ASGIRoute, Dict[HttpMethod, HTTPRouteHandler]]]:
        """
        Maps route handlers to http methods
        """
        handlers_map: Dict[str, Any] = {}
        if isinstance(value, BaseRouteHandler):
            for path in value.paths:
                if isinstance(value, HTTPRouteHandler):
                    handlers_map[path] = {http_method: value for http_method in value.http_methods}
                elif isinstance(value, (WebsocketRouteHandler, ASGIRouteHandler)):
                    handlers_map[path] = value
        elif isinstance(value, Router):
            handlers_map = value.route_handler_method_map
        else:
            # we are dealing with a controller
            for route_handler in value.get_route_handlers():
                for handler_path in route_handler.paths:
                    path = join_paths([value.path, handler_path]) if handler_path else value.path
                    if isinstance(route_handler, HTTPRouteHandler):
                        if not isinstance(handlers_map.get(path), dict):
                            handlers_map[path] = {}
                        for http_method in route_handler.http_methods:
                            handlers_map[path][http_method] = route_handler
                    else:
                        handlers_map[path] = cast(Union[WebsocketRouteHandler, ASGIRouteHandler], route_handler)
        return handlers_map.items()

    def validate_registration_value(
        self, value: ControllerRouterHandler
    ) -> Union[Controller, BaseRouteHandler, "Router"]:
        """
        Validates that the value passed to the register method is supported
        """
        if isclass(value) and issubclass(cast(Type[Controller], value), Controller):
            return cast(Type[Controller], value)(owner=self)
        if not isinstance(value, (Router, BaseRouteHandler)):
            raise ImproperlyConfiguredException(
                "Unsupported value passed to `Router.register`. "
                "If you passed in a function or method, "
                "make sure to decorate it first with one of the routing decorators"
            )
        if isinstance(value, Router):
            if value.owner:
                raise ImproperlyConfiguredException(f"Router with path {value.path} has already been registered")
            if value is self:
                raise ImproperlyConfiguredException("Cannot register a router on itself")
        value.owner = self
        return cast(Union[Controller, BaseRouteHandler, "Router"], value)

    def register(self, value: ControllerRouterHandler) -> List[BaseRoute]:
        """
        Register a Controller, Route instance or RouteHandler on the router

        Accepts a subclass or instance of Controller, an instance of Router or a function/method that has been decorated
        by any of the routing decorators (e.g. route, get, post...) exported from 'starlite.routing'
        """
        validated_value = self.validate_registration_value(value)
        routes: List[BaseRoute] = []
        for route_path, handler_or_method_map in self.map_route_handlers(value=validated_value):
            path = join_paths([self.path, route_path])
            if isinstance(handler_or_method_map, WebsocketRouteHandler):
                route: BaseRoute = WebSocketRoute(path=path, route_handler=handler_or_method_map)
                self.routes.append(route)
            elif isinstance(handler_or_method_map, ASGIRouteHandler):
                route = ASGIRoute(path=path, route_handler=handler_or_method_map)
                self.routes.append(route)
            else:
                existing_handlers: List[HTTPRouteHandler] = list(self.route_handler_method_map.get(path, {}).values())  # type: ignore
                route_handlers = unique(list(cast(Dict[HttpMethod, HTTPRouteHandler], handler_or_method_map).values()))
                if existing_handlers:
                    route_handlers.extend(unique(existing_handlers))
                    existing_route_index = find_index(
                        self.routes, lambda x: x.path == path  # pylint: disable=cell-var-from-loop
                    )
                    assert existing_route_index != -1, "unable to find_index existing route index"
                    route = HTTPRoute(
                        path=path,
                        route_handlers=route_handlers,
                    )
                    self.routes[existing_route_index] = route
                else:
                    route = HTTPRoute(path=path, route_handlers=route_handlers)
                    self.routes.append(route)
            routes.append(route)
        return routes



from inspect import Signature, iscoroutinefunction
from typing import cast

from pydantic.typing import AnyCallable

from starlite.exceptions import ImproperlyConfiguredException
from starlite.handlers.base import BaseRouteHandler
from starlite.types import AsyncAnyCallable


class WebsocketRouteHandler(BaseRouteHandler):
    def __call__(self, fn: AsyncAnyCallable) -> "WebsocketRouteHandler":
        """
        Replaces a function with itself
        """
        self.fn = fn
        self.validate_handler_function()
        return self

    def validate_handler_function(self) -> None:
        """
        Validates the route handler function once it's set by inspecting its return annotations
        """
        super().validate_handler_function()
        signature = Signature.from_callable(cast(AnyCallable, self.fn))

        if signature.return_annotation is not None:
            raise ImproperlyConfiguredException("Websocket handler functions should return 'None'")
        if "socket" not in signature.parameters:
            raise ImproperlyConfiguredException("Websocket handlers must set a 'socket' kwarg")
        if "request" in signature.parameters:
            raise ImproperlyConfiguredException("The 'request' kwarg is not supported with websocket handlers")
        if "data" in signature.parameters:
            raise ImproperlyConfiguredException("The 'data' kwarg is not supported with websocket handlers")
        if not iscoroutinefunction(self.fn) and not iscoroutinefunction(self.fn.__call__):  # type: ignore[operator]
            raise ImproperlyConfiguredException("Functions decorated with 'websocket' must be async functions")


websocket = WebsocketRouteHandler



from atexit import register
from logging.config import ConvertingList  # type: ignore
from logging.handlers import QueueHandler, QueueListener
from queue import Queue
from typing import List


def _resolve_handlers(handlers: List[str]) -> List:
    """
    Converts list of string of handlers to the object of respective handler.
    Indexing the list performs the evaluation of the object.
    """
    if not isinstance(handlers, ConvertingList):
        return handlers

    return [handlers[i] for i in range(len(handlers))]


class QueueListenerHandler(QueueHandler):
    """
    Configures queue listener and handler to support non-blocking logging configuration.
    """

    def __init__(
        self, handlers: List, respect_handler_level: bool = False, auto_run: bool = True, queue: Queue = Queue(-1)
    ):
        super().__init__(queue)
        self.handlers = _resolve_handlers(handlers)
        self._listener: QueueListener = QueueListener(
            self.queue, *self.handlers, respect_handler_level=respect_handler_level
        )
        if auto_run:
            self.start()
            register(self.stop)

    def start(self) -> None:
        """
        starts the listener.
        """
        self._listener.start()

    def stop(self) -> None:
        """
        Manually stop a listener.
        """
        self._listener.stop()



from http import HTTPStatus
from typing import Any, Dict, Optional

from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_405_METHOD_NOT_ALLOWED,
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_503_SERVICE_UNAVAILABLE,
)


class StarLiteException(Exception):
    def __init__(self, detail: str = ""):
        self.detail = detail
        super().__init__()

    def __repr__(self) -> str:
        if self.detail:
            return f"{self.__class__.__name__} - {self.detail}"
        return self.__class__.__name__


class MissingDependencyException(StarLiteException, ImportError):
    pass


class HTTPException(StarLiteException, StarletteHTTPException):
    status_code = HTTP_500_INTERNAL_SERVER_ERROR
    extra: Optional[Dict[str, Any]] = None

    def __init__(  # pylint: disable=super-init-not-called
        self, detail: Optional[str] = None, status_code: Optional[int] = None, extra: Optional[Dict[str, Any]] = None
    ):
        if status_code:
            self.status_code = status_code
        self.detail = detail or HTTPStatus(self.status_code).phrase
        self.extra = extra

    def __repr__(self) -> str:
        return f"{self.status_code} - {self.__class__.__name__} - {self.detail}"


class ImproperlyConfiguredException(HTTPException, ValueError):
    pass


class ValidationException(HTTPException, ValueError):
    status_code = HTTP_400_BAD_REQUEST


class NotAuthorizedException(HTTPException):
    status_code = HTTP_401_UNAUTHORIZED


class PermissionDeniedException(HTTPException):
    status_code = HTTP_403_FORBIDDEN


class NotFoundException(HTTPException, ValueError):
    status_code = HTTP_404_NOT_FOUND


class MethodNotAllowedException(HTTPException):
    status_code = HTTP_405_METHOD_NOT_ALLOWED


class InternalServerException(HTTPException):
    status_code = HTTP_500_INTERNAL_SERVER_ERROR


class ServiceUnavailableException(HTTPException):
    status_code = HTTP_503_SERVICE_UNAVAILABLE


class TemplateNotFound(InternalServerException):
    def __init__(self, template_name: str):
        super().__init__(detail=f"Template {template_name} not found.")



from starlite.datastructures import File, Redirect, State, Stream, Template

from .app import Starlite
from .config import (
    CacheConfig,
    CORSConfig,
    OpenAPIConfig,
    StaticFilesConfig,
    TemplateConfig,
)
from .connection import Request, WebSocket
from .controller import Controller
from .dto import DTOFactory
from .enums import (
    HttpMethod,
    MediaType,
    OpenAPIMediaType,
    RequestEncodingType,
    ScopeType,
)
from .exceptions import (
    HTTPException,
    ImproperlyConfiguredException,
    InternalServerException,
    MissingDependencyException,
    NotAuthorizedException,
    NotFoundException,
    PermissionDeniedException,
    ServiceUnavailableException,
    StarLiteException,
    ValidationException,
)
from .handlers import (
    ASGIRouteHandler,
    BaseRouteHandler,
    HTTPRouteHandler,
    WebsocketRouteHandler,
    asgi,
    delete,
    get,
    patch,
    post,
    put,
    route,
    websocket,
)
from .logging import LoggingConfig
from .middleware import AbstractAuthenticationMiddleware, AuthenticationResult
from .openapi.controller import OpenAPIController
from .params import Body, Parameter
from .plugins import PluginProtocol
from .provide import Provide
from .queue_listener_handler import QueueListenerHandler
from .response import Response
from .router import Router
from .routes import BaseRoute, HTTPRoute, WebSocketRoute
from .testing import TestClient, create_test_client, create_test_request
from .types import MiddlewareProtocol, Partial, ResponseHeader

__all__ = [
    "ASGIRouteHandler",
    "AbstractAuthenticationMiddleware",
    "AuthenticationResult",
    "BaseRoute",
    "BaseRouteHandler",
    "Body",
    "CacheConfig",
    "CORSConfig",
    "Controller",
    "DTOFactory",
    "File",
    "HTTPException",
    "HTTPRoute",
    "HTTPRouteHandler",
    "HttpMethod",
    "ImproperlyConfiguredException",
    "InternalServerException",
    "LoggingConfig",
    "MediaType",
    "MiddlewareProtocol",
    "MissingDependencyException",
    "NotAuthorizedException",
    "NotFoundException",
    "OpenAPIConfig",
    "OpenAPIController",
    "OpenAPIMediaType",
    "Parameter",
    "Partial",
    "PermissionDeniedException",
    "PluginProtocol",
    "Provide",
    "Redirect",
    "Request",
    "RequestEncodingType",
    "Response",
    "ResponseHeader",
    "Router",
    "ScopeType",
    "ServiceUnavailableException",
    "StarLiteException",
    "Starlite",
    "State",
    "StaticFilesConfig",
    "Stream",
    "Template",
    "TemplateConfig",
    "TestClient",
    "QueueListenerHandler",
    "ValidationException",
    "WebSocket",
    "WebSocketRoute",
    "WebsocketRouteHandler",
    "asgi",
    "create_test_client",
    "create_test_request",
    "delete",
    "get",
    "patch",
    "post",
    "put",
    "route",
    "websocket",
]
